#import squidpy as sq
import scanpy as sc
import anndata as ad
import numpy as np
import pandas as pd
import squidpy as sq
import matplotlib.pyplot as plt
import scipy
import random 
import matplotlib
import seaborn as sns
import math
#import time
from scipy.special import softmax
from scipy.optimize import minimize
from numpy.random import uniform
from sklearn.preprocessing import OneHotEncoder
from sklearn.preprocessing import LabelEncoder
from .utils import *
from .Annealing import *

import numba




def mutate_cell(celltype_assignment = None, mutate_ct = None, mutate_toct = None,times = None):
    
    mutate_index = np.random.choice(np.where(celltype_assignment==mutate_ct)[0],size = times, replace=False )
    
    celltype_assignment[mutate_index] = mutate_toct
    
    return celltype_assignment
    

def mutate(celltype_assignment=None, Num_ct_sample=None):
    compare = np.bincount(celltype_assignment) - Num_ct_sample
    while compare.any(): 
        print("np.bincount(celltype_assignment):  ",np.bincount(celltype_assignment))
        print("Num_ct_sample: ",Num_ct_sample)
        print("compare: ",compare)
        mutate_ct = np.where(compare > 0)[0][0]
        mutate_toct = np.where(compare < 0)[0][0]
        celltype_assignment = mutate_cell(
            celltype_assignment=celltype_assignment,
            mutate_ct=mutate_ct,
            mutate_toct=mutate_toct,
            times=np.min([compare[mutate_ct], -compare[mutate_toct]]))
        compare = np.bincount(celltype_assignment) - Num_ct_sample
    return celltype_assignment

def gen_lowpixel_grid(Num_sample=None,
                       original_grid = None,
                      grid_row=None,
                      grid_col=None,
                      windows_row=None,
                      windows_col=None):
    '''
    Attention:
        original_grid is grid location in original resolution (n*2)
        when `original_grid` is not None, you must input grid_row and grid_col 
            grid_row : the number of grid by row
            grid_col : the number of grid by col
        If you don't have original_grid, you must input Num_sample
        Here, cells are arranged into a square grid according to the sample size,
        and the length of its sides is `np.sqrt(Num_sample).astype(int)`
    
    '''
    if grid_row is None and grid_col is None:
        grid_row = np.sqrt(Num_sample).astype(int)
        grid_col = grid_row
        
    a, b = np.meshgrid(np.arange(grid_row),
                       np.arange(grid_col),
                       indexing='ij')

    lra = a[np.arange(0, grid_row, windows_row
                      ), :len(np.arange(0, grid_col, windows_col))].reshape(-1)
    lrb = b[:len(np.arange(0, grid_row, windows_row)),
            np.arange(0, grid_col, windows_col)].reshape(-1)
    grid_row_num, grid_col_num = grid_row, grid_col
    lr_spatial_idx = np.vstack((lra, lrb)).T
    
    if original_grid is None:
        lr_spatial = lr_spatial_idx.copy() 
    else:
        original_grid_x = original_grid[:,0].reshape(grid_row,grid_col)
        original_grid_y = original_grid[:,1].reshape(grid_row,grid_col)
        lr_spatial_x = original_grid_x[lra,lrb].reshape(-1)
        lr_spatial_y = original_grid_y[lra,lrb].reshape(-1)
        lr_spatial = np.vstack((lr_spatial_x, lr_spatial_y)).T
    return lr_spatial_idx, lr_spatial, grid_row_num, grid_col_num



def enhance_res(Num_sample=None,
                original_grid=None,
                grid_row=None,
                grid_col=None,
                windows_row=None,
                windows_col=None,
                target_trans=None,
                Num_celltype=None,
                prior=None,
                T=None,
                lrct_assignment=None,
                swap_num=None,
                tol=None,
                smallsample_max_iter=None,
                bigsample_max_iter=None):
    '''
      如果只给定 Num_sample 那么默认是正方形网格，如果给定了grid_row and grid_col，则按照
      该方式生成。要注意Num_sample = grid_row * grid_col
      windows_row: 低分辨率下，划分网格的长度
      windows_col: 低分辨率下，划分网格的宽度
      T: 退火的初始温度
      lrct_assignment: 低分辨率下的网格点的细胞类型分配，注意是一个shape为grid_row,grid_col的矩阵
    '''

    ##选取低分率下的网格代表元
    lr_spatial_idx, lr_spatial, grid_row_num, grid_col_num = gen_lowpixel_grid(
	 original_grid=original_grid,
        Num_sample=Num_sample,
        grid_row=grid_row,
        grid_col=grid_col,
        windows_row=windows_row,
        windows_col=windows_col)

    
    if lrct_assignment is None:
        Num_ct_sample = get_ct_sample(Num_celltype=Num_celltype,
                                      Num_sample=lr_spatial.shape[0],
                                      prior=prior)
        celltype_assignment = init_ct(Num_celltype=Num_celltype,
                                      Num_ct_sample=Num_ct_sample)
        #print("lrct_None_Num_ct_sample:  ",Num_ct_sample)

    else:
        celltype_assignment = lrct_assignment[lr_spatial_idx[:, 0],
                                              lr_spatial_idx[:, 1]].reshape(-1)

    ## 要能整除！
    
    ##最后一次的时候，要考虑此时的ct比例，要与prior一致，所以要mutate
    if windows_row == 1 and windows_col == 1:
        Num_ct_sample = get_ct_sample(Num_celltype=Num_celltype,
                                      Num_sample=Num_sample,
                                      prior=prior)
        celltype_assignment = mutate(celltype_assignment=celltype_assignment,
                                     Num_ct_sample=Num_ct_sample)

    sn = get_spaital_network(Num_sample=lr_spatial.shape[0],
                             spatial=lr_spatial,
                             n_neighs=8)
    onehot_ct = get_onehot_ct(init_assign=celltype_assignment)
    nb_count = np.array(sn * onehot_ct, dtype=np.float32)

    resct_assignment = STsim(Num_sample=lr_spatial.shape[0],
                             Num_celltype=Num_celltype,
                             celltype_assignment=celltype_assignment,
                             target_trans=target_trans,
                             T=T,
                             chain_len=100,
                             error=1e2,
                             onehot_ct=onehot_ct,
                             tol=tol,
                             decay=0.5,
                             nb_count=nb_count,
                             sn=sn,
                             swap_num=swap_num,
                             smallsample_max_iter=smallsample_max_iter,
                             bigsample_max_iter=bigsample_max_iter)
    print("grid_row:",grid_row_num)
    print("windows_row:",windows_row)

    ct_grid = np.array(resct_assignment).reshape(
        len(np.arange(0, grid_row_num, windows_row)),
        len(np.arange(0, grid_col_num, windows_col)))
    ct_grid = np.repeat(np.repeat(ct_grid, windows_row, axis=0),
                        windows_col,
                        axis=1)

    return ct_grid, lr_spatial

def enhance_loop(Num_sample=None,
                 Num_celltype=None,
                 prior=None,
                 target_trans=None,
                 original_grid=None,
                 celltype_assignment = None,
                 grid_row=None,
                 grid_col=None,
                 loop_times=None,
                 windows_row_list=None,
                 windows_col_list=None,
                 swap_num_list=None,
                 tol_list=None,
                 T=1e-5,
                 smallsample_max_iter=80000,
                 bigsample_max_iter=10000):
    '''
    Attention: 我们可以基于样本量推荐一个loop_times,
                小样本许多参数是固定的，you can see `STsim` function
                验证用户填写的list是否符合规则
                
    
    '''
    if Num_sample <=10000: loop_times  = 1
    if loop_times is None:
        loop_times = np.round(np.log2(Num_sample / 1000) / 2).astype(int) + 1

    ##需要注意 不同分辨率窗口的row 与col 彼此之间要整除
    if windows_col_list is None and windows_row_list is None:
        windows_row_list = [2**i for i in range(loop_times)]
        windows_row_list.reverse()
        windows_col_list = [2**i for i in range(loop_times)]
        windows_col_list.reverse()

    ## 默认小于10000，swap_num = 1，并且要与windows对应
    if swap_num_list is None:
        swap_num_list = [2**i for i in range(1, loop_times + 1)]

    if tol_list is None:
        tol_list = [2e-2] * loop_times
    
   # celltype_assignment = None
    for i in np.arange(loop_times):
        celltype_assignment, lr_spatial = enhance_res(
                Num_sample=Num_sample,
                original_grid=original_grid,
                grid_row=grid_row,
                grid_col=grid_col,
                windows_row=windows_row_list[i],
                windows_col=windows_col_list[i],
                target_trans=target_trans,
                Num_celltype=Num_celltype,
                prior=prior,
                T=1e-5,
                lrct_assignment=celltype_assignment,
                swap_num=swap_num_list[i],
                tol=tol_list[i],
                smallsample_max_iter=smallsample_max_iter,
                bigsample_max_iter=bigsample_max_iter)
    return celltype_assignment, lr_spatial, loop_times


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sun Mar  8 19:43:29 2020

@author: heavens
"""
import numpy as np
from sklearn.decomposition import PCA
from scipy.sparse import dok_matrix
import pickle
from sklearn import manifold
import os 

def tag2int(label):
    tags = list(set(label))
    int_label = [tags.index(x) for x in label]
    return np.asarray(int_label),tags

def one_hot_vector(label,class_n = None):
    label_n = len(label)
    if label.dtype != np.int:
        print("Input vector has a dtype %s"%(label.dtype))
        print("Transfering the data type into int.")
        _,tags = tag2int(label)
    else:
        tags = list(set(label))
    if class_n is None:
        class_n = len(tags)
    one_hot = np.zeros((label_n,class_n))
    for tag_idx,tag in enumerate(tags):
        one_hot[label==tag,tag_idx] = 1
    return one_hot,tags

def embedding_reduce(X,embedding):
    reduced_X = X@embedding
    return reduced_X,embedding

def tsne_reduce(X,dims = 2,method = "exact",tsne = None):
    if tsne is None:
        tsne = manifold.TSNE(n_components=dims, 
                             init='pca', 
                             random_state=0,
                             method = method)
    reduced_X = tsne.fit_transform(X)
    return reduced_X,tsne

def pca_reduce(X, dims=2,pca = None):
    """ Reduce the dimensions of X down to dims using PCA
    X has shape (n, d)
    Returns: The reduced X of shape (n, dims)
    		 The fitted PCA model used to reduce X
    """
    print("reducing dimensions using PCA")
    X = X - np.mean(X,axis = 0,keepdims = True)
    std = np.std(X,axis = 0, keepdims = True)
    std[std==0] += 1e-9 #Add a small pseudo count to avoid divide by 0 error.
    X = X/std
    if pca is None:
        pca = PCA(n_components = dims)
        pca.fit(X)
    X_reduced = pca.transform(X)
    return X_reduced,pca

def KL_divergence(prob1,prob2,pesudo=1e-6):
    if 0 in prob1:
        prob1 = prob1+pesudo
        prob1 = prob1/np.sum(prob1)
        print("Zero element in the first distribution, add pesudo count %.2f."%(pesudo))
    if 0 in prob2:
        prob2 = prob2+pesudo
        prob2 = prob2/np.sum(prob2)
        print("Zero element in the second distribution, add pesudo count %.2f."%(pesudo))
    kld = np.sum([x*np.log(x/y) for x,y in zip(prob1,prob2)])
    return kld

def get_adjacency(coordinate,
                  threshold_distance,
                  exclude_self = False):
    sample_n = coordinate.shape[0]
    adjacency = dok_matrix((sample_n,sample_n),bool)
    for i in np.arange(sample_n):
        difference = coordinate - coordinate[i]
        difference = difference**2
        euclidean_distance = np.sqrt(np.sum(difference,axis = 1))
        adjacency_array = euclidean_distance<threshold_distance
        if exclude_self:
            adjacency_array[i] = False
        cols = np.where(adjacency_array)[0]
        for col in cols:
            adjacency[i,col] = True
    return adjacency

def get_adjacency_knearest(coordinate,
                           nearest_k,
                           exclude_self = False):
    sample_n = coordinate.shape[0]
    adjacency = dok_matrix((sample_n,sample_n),bool)
    if exclude_self:
        nearest_k +=1
    for i in np.arange(sample_n):
        difference = coordinate - coordinate[i]
        euclidean_distance = np.sqrt(difference[:,0]**2+difference[:,1]**2)
        sort_idx = np.argsort(euclidean_distance)
        for col in sort_idx[:nearest_k]:
            adjacency[i,col] = True
        if exclude_self:
            adjacency[i,i] = False
    return adjacency

def get_knearest_distance(coordinate,
                          nearest_k,
                          exclude_self = False):
    sample_n = coordinate.shape[0]
    if exclude_self:
        nearest_k +=1
    distance_list = []
    for i in np.arange(sample_n):
        difference = coordinate - coordinate[i]
        euclidean_distance = np.sqrt(difference[:,0]**2+difference[:,1]**2)
        sort_distance = np.sort(euclidean_distance)
        distance_list.append(sort_distance[nearest_k])
    return distance_list    

def get_neighbourhood_count(adjacency, 
                            label,
                            exclude_self = True,
                            one_hot_label = True):
    """Get the neighbourhood type count given a adjacency matrix and label
    Args:
        adjacency: A N-by-N 0-1 adjacency matrix, N is the number of samples.
        label: A length N vector indicate the label tags.
        exclude_self: A boolean variable, if true the self-count is excluded.
    Return:
        nb_count: A N-by-M matrix given the neighbourhood count, where N is the
            number of sample, and M is the number of types of label.
    """
    sample_n = adjacency.shape[0]
    assert label.shape[0] == sample_n
    tags = np.unique(label)
    if one_hot_label:
        one_hot = label
    else:
        label_n = len(tags)
        one_hot = np.zeros((sample_n,label_n))
        one_hot[np.arange(sample_n),label] = 1
    if exclude_self:
        diag_mask = np.eye(sample_n).astype(bool)
        adjacency[diag_mask] = 0
    nb_count = adjacency.dot(one_hot)
    return nb_count

class DataLoader():
    def __init__(self,
                 xs,
                 y = None,
                 for_eval = False):
        """Base class for loading the data.
        Input Args:
            xs: The input data, can be a tuple of arrays with the same length in
                0 dimension.
            y: The label, need to have same length as x.
            for_eval: If the dataset is for evaluation.
        """
        if type(xs) is tuple:
            self.xs = xs
        else:
            self.xs = tuple(xs)
        self.y = y
        self.epochs_completed = 0
        self._index_in_epoch = 0
        self.sample_n = self.xs[0].shape[0]
        for x in self.xs:
            assert x.shape[0] == self.sample_n
        self._perm = np.arange(self.sample_n)
        self.for_eval = for_eval
        
    def read_into_memory(self, index):
        if self.for_eval:
            batch_y = None
        else:
            batch_y = self.y[index]
        batch_x = tuple([x[index] for x in self.xs])
        return batch_x,batch_y
    def _multi_concatenate(self,input_tuple,axis=0):
        result = [np.concatenate(x,axis = axis) for x in zip(*input_tuple)]
        return tuple(result)
    def next_batch(self, batch_size, shuffle=True):
        """Return next batch in batch_size from the data set.
            Input Args:
                batch_size:A scalar indicate the batch size.
                shuffle: boolean, indicate if the data should be shuffled after each epoch.
            Output Args:
                inputX,sequence_length,label_batch: tuple of (indx,vals,shape)
        """
        if self.epochs_completed>=1 and self.for_eval:
            print("Warning, evaluation dataset already finish one iteration.")
        start = self._index_in_epoch
        # Shuffle for the first epoch
        if self.epochs_completed == 0 and start == 0:
            if shuffle:
                np.random.shuffle(self._perm)
        # Go to the next epoch
        if start + batch_size >= self.sample_n:
            # Finished epoch
            self.epochs_completed += 1
            # Get the rest samples in this epoch
            rest_sample_n = self.sample_n - start
            x_rest_part, y_rest_part = self.read_into_memory(self._perm[start:self.sample_n])
            start = 0
            if self.for_eval:
                x_batch = x_rest_part
                y_batch = y_rest_part
                self._index_in_epoch = 0
                end = 0
            # Shuffle the data
            else:
                if shuffle:
                    np.random.shuffle(self._perm)
                # Start next epoch
                self._index_in_epoch = batch_size - rest_sample_n
                end = self._index_in_epoch
                x_new_part, y_new_part = self.read_into_memory(
                    self._perm[start:end])
                if x_rest_part[0].shape[0] == 0:
                    x_batch = x_new_part
                    y_batch = y_new_part
                elif x_new_part[0].shape[0] == 0:
                    x_batch = x_rest_part
                    y_batch = y_rest_part
                else:
                    x_batch = self._multi_concatenate((x_rest_part, x_new_part))
                    y_batch = np.concatenate((y_rest_part, y_new_part),axis = 0)
        else:
            self._index_in_epoch += batch_size
            end = self._index_in_epoch
            x_batch,y_batch = self.read_into_memory(
                self._perm[start:end])
        return x_batch,y_batch
    
def save_loader(loader,save_f):
    with open(save_f,'wb+') as f:
        pickle.dump(loader,f)

def save_smfish(loader,save_prefix,is_labeled=False):
    """Save the data in the loader as the smfishHmrf format:
    http://spatial.rc.fas.harvard.edu/install.html
    """
    norm_expr = loader.gene_expression - np.mean(loader.gene_expression,axis = 1,keepdims = True)
    std_expr = np.std(loader.gene_expression,axis = 1,keepdims = True)
    zero_mask = std_expr != 0
    zero_mask = np.reshape(zero_mask,newshape=zero_mask.shape[0])
    norm_expr = norm_expr[zero_mask,:] / std_expr[zero_mask]
    with open(save_prefix+'.expression','w+') as f:
        for i,expr in enumerate(norm_expr):
            f.write("%d %s\n"%(i+1,' '.join([str(e) for e in expr])))
    coordinate = loader.coordinate[zero_mask,:]
    with open(save_prefix+'.coordinates','w+') as f:
        if coordinate.shape[1]==2:
            for i,p in enumerate(coordinate):
                f.write("%d %.3f %.3f %.3f\n"%(i+1,0,p[0],p[1]))
        elif coordinate.shape[1] ==3:
            for i,p in enumerate(coordinate):
                f.write("%d %.3f %.3f %.3f\n"%(i+1,p[2],p[0],p[1]))
    with open(save_prefix+'.genes','w+') as f:
        for i,g in enumerate(loader.gene_list):
            f.write("%s\n"%(g))
    if is_labeled:
        with open(save_prefix+'.labels','w+') as f:
            for i,c in enumerate(loader.cell_labels):
                f.write("%d %d\n"%(i+1,c))


def load_loader(save_f):
    with open(save_f,'rb') as f:
        loader = pickle.load(f)
    return loader


import argparse
import os, sys
from .scsim import scsim
import numpy as np
import time
import pandas as pd


def save_df(obj, filename):
    '''Save pandas dataframe in compressed format'''
    np.savez_compressed(filename, data=obj.values, index=obj.index.values, columns=obj.columns.values)


def load_df(filename):
    with np.load(filename, allow_pickle=True) as f:
        obj = pd.DataFrame(**f)
    return obj
    
    
# def parse_args():
#     parser = argparse.ArgumentParser(description='Run scsim with specified input arguments')
#     parser.add_argument('--outdir', type=str, default='scsim-%s-%s-%s-%s-%s-%s-%s-%s',
#                         help='Output directory base')
#     parser.add_argument('--seed', type=int, help='simulation seed')
#     parser.add_argument('--numsims', type=int, help='number of sims to run',
#                         default=20)
#     parser.add_argument('--deloc', type=float,
#                         help='devalue',
#                         default=1.)
#     parser.add_argument('--K', type=int,
#                         help='Number of identity programs',
#                         default=10)
#     parser.add_argument('--nproggoups', type=int,
#                         help='Number of groups expressing activity program. Default is 1/3 of K rounded down',
#                         default=None)
#     parser.add_argument('--ncells', type=int,
#                         help='Total number of cells',
#                         default=10000)
#     parser.add_argument('--doubletfrac', type=float,
#                         help='Percentage of doublet cells',
#                         default=0.)
#     a = parser.parse_args()
#     return(a.outdir, a.seed, a.numsims, a.deloc, a.K, a.nproggoups, a.ncells, a.doubletfrac)

def run_scsim(outdir ='scsim-%s-%s-%s-%s-%s-%s-%s-%s',randseed = None,numsims = 20,
         deval = 1.,K =10, nproggroups = None, ncells = 10000, doubletfrac = 0. ):
    
#     def parse_args():
#     parser = argparse.ArgumentParser(description='Run scsim with specified input arguments')
#     parser.add_argument('--outdir', type=str, default='scsim-%s-%s-%s-%s-%s-%s-%s-%s',
#                         help='Output directory base')
#     parser.add_argument('--seed', type=int, help='simulation seed')
#     parser.add_argument('--numsims', type=int, help='number of sims to run',
#                         default=20)
#     parser.add_argument('--deloc', type=float,
#                         help='devalue',
#                         default=1.)
#     parser.add_argument('--K', type=int,
#                         help='Number of identity programs',
#                         default=10)
#     parser.add_argument('--nproggoups', type=int,
#                         help='Number of groups expressing activity program. Default is 1/3 of K rounded down',
#                         default=None)
#     parser.add_argument('--ncells', type=int,
#                         help='Total number of cells',
#                         default=10000)
#     parser.add_argument('--doubletfrac', type=float,
#                         help='Percentage of doublet cells',
#                         default=0.)
#     a = parser.parse_args()
    
#     (outdir, randseed, numsims, deval, K, nproggroups, ncells, doubletfrac) = parse_args()
    ngenes=10000
    nproggenes = 400
    ndoublets=int(doubletfrac*ncells)
    
    deloc=deval
    progdeloc=deval
    descale=1.0
    progcellfrac = .35
    deprob = .025

    if nproggroups is None:
        nproggroups = int(K/3)
        
    proggroups = list(range(1, nproggroups+1))

    simulator = scsim(ngenes=ngenes, ncells=ncells, ngroups=K, libloc=7.64, libscale=0.78,
                 mean_rate=7.68,mean_shape=0.34, expoutprob=0.00286,
                 expoutloc=6.15, expoutscale=0.49,
                 diffexpprob=deprob, diffexpdownprob=0., diffexploc=deloc, diffexpscale=descale,
                 bcv_dispersion=0.448, bcv_dof=22.087, ndoublets=ndoublets,
                 nproggenes=nproggenes, progdownprob=0., progdeloc=progdeloc,
                 progdescale=descale, progcellfrac=progcellfrac, proggoups=proggroups,
                 minprogusage=.1, maxprogusage=.7, seed=randseed)


    start_time = time.time()
    simulator.simulate()
    end_time = time.time()
    print('%.3f minutes elapsed for seed %d' % ((end_time-start_time)/60, randseed))

    save_df(simulator.cellparams, os.path.join(outdir, 'cellparams'))
    save_df(simulator.geneparams, os.path.join(outdir, 'geneparams'))
    save_df(simulator.counts, os.path.join(outdir, 'counts'))








import pandas as pd
import numpy as np
import sys

class scsim:
    def __init__(self, ngenes=10000, ncells=100, seed=757578,
                 mean_rate=.3, mean_shape=.6, libloc=11, libscale=0.2,
                expoutprob=.05, expoutloc=4, expoutscale=0.5, ngroups=1,
                diffexpprob=.1, diffexpdownprob=.5,
                diffexploc=.1, diffexpscale=.4, bcv_dispersion=.1,
                bcv_dof=60, ndoublets=0, groupprob=None,
                nproggenes=None, progdownprob=None, progdeloc=None,
                progdescale=None, proggoups=None, progcellfrac=None,
                minprogusage=.2, maxprogusage=.8):

        self.ngenes = ngenes
        self.ncells = ncells
        self.seed = seed
        self.mean_rate = mean_rate
        self.mean_shape = mean_shape
        self.libloc = libloc
        self.libscale = libscale
        self.expoutprob = expoutprob
        self.expoutloc = expoutloc
        self.expoutscale = expoutscale
        self.ngroups = ngroups
        self.diffexpprob = diffexpprob
        self.diffexpdownprob = diffexpdownprob
        self.diffexploc = diffexploc
        self.diffexpscale = diffexpscale
        self.bcv_dispersion = bcv_dispersion
        self.bcv_dof = bcv_dof
        self.ndoublets = ndoublets
        self.init_ncells = ncells+ndoublets
        self.nproggenes=nproggenes
        self.progdownprob=progdownprob
        self.progdeloc=progdeloc
        self.progdescale=progdescale
        self.proggoups=proggoups
        self.progcellfrac = progcellfrac
        self.minprogusage = minprogusage
        self.maxprogusage = maxprogusage

        if groupprob is None:
            self.groupprob = [1/float(self.ngroups)]*self.ngroups
        elif (len(groupprob) == self.ngroups) & (np.abs(np.sum(groupprob) - 1) < (10**-6)):
            self.groupprob = groupprob
        else:
            sys.exit('Invalid groupprob input')


    def simulate(self):
        np.random.seed(self.seed)
        print('Simulating cells')
        self.cellparams = self.get_cell_params()
        print('Simulating gene params')
        self.geneparams = self.get_gene_params()

        if (self.nproggenes is not None) and (self.nproggenes > 0):
            print('Simulating program')
            self.simulate_program()

        print('Simulating DE')
        self.sim_group_DE()

        print('Simulating cell-gene means')
        self.cellgenemean = self.get_cell_gene_means()
        if self.ndoublets > 0:
            print('Simulating doublets')
            self.simulate_doublets()

        print('Adjusting means')
        self.adjust_means_bcv()
        print('Simulating counts')
        self.simulate_counts()

    def simulate_counts(self):
        '''Sample read counts for each gene x cell from Poisson distribution
        using the variance-trend adjusted updatedmean value'''
        self.counts = pd.DataFrame(np.random.poisson(lam=self.updatedmean),
                                   index=self.cellnames, columns=self.genenames)

    def adjust_means_bcv(self):
        '''Adjust cellgenemean to follow a mean-variance trend relationship'''
        self.bcv = self.bcv_dispersion + (1 / np.sqrt(self.cellgenemean))
        chisamp = np.random.chisquare(self.bcv_dof, size=self.ngenes)
        self.bcv = self.bcv*np.sqrt(self.bcv_dof / chisamp)
        self.updatedmean = np.random.gamma(shape=1/(self.bcv**2),
                                           scale=self.cellgenemean*(self.bcv**2))
        self.bcv = pd.DataFrame(self.bcv, index=self.cellnames, columns=self.genenames)
        self.updatedmean = pd.DataFrame(self.updatedmean, index=self.cellnames,
                                        columns=self.genenames)


    def simulate_doublets(self):
        ## Select doublet cells and determine the second cell to merge with
        d_ind = sorted(np.random.choice(self.ncells, self.ndoublets,
                                        replace=False))
        d_ind = ['Cell%d' % (x+1) for x in d_ind]
        self.cellparams['is_doublet'] = False
        self.cellparams.loc[d_ind, 'is_doublet'] = True
        extraind = self.cellparams.index[-self.ndoublets:]
        group2 = self.cellparams.ix[extraind, 'group'].values
        self.cellparams['group2'] = -1
        self.cellparams.loc[d_ind, 'group2'] = group2

        ## update the cell-gene means for the doublets while preserving the
        ## same library size
        dmean = self.cellgenemean.loc[d_ind,:].values
        dmultiplier = .5 / dmean.sum(axis=1)
        dmean = np.multiply(dmean, dmultiplier[:, np.newaxis])
        omean = self.cellgenemean.loc[extraind,:].values
        omultiplier = .5 / omean.sum(axis=1)
        omean = np.multiply(omean, omultiplier[:,np.newaxis])
        newmean = dmean + omean
        libsize = self.cellparams.loc[d_ind, 'libsize'].values
        newmean = np.multiply(newmean, libsize[:,np.newaxis])
        self.cellgenemean.loc[d_ind,:] = newmean
        ## remove extra doublet cells from the data structures
        self.cellgenemean.drop(extraind, axis=0, inplace=True)
        self.cellparams.drop(extraind, axis=0, inplace=True)
        self.cellnames = self.cellnames[0:self.ncells]


    def get_cell_gene_means(self):
        '''Calculate each gene's mean expression for each cell while adjusting
        for the library size'''


        group_genemean = self.geneparams.loc[:,[x for x in self.geneparams.columns if ('_genemean' in x) and ('group' in x)]].T.astype(float)
        group_genemean = group_genemean.div(group_genemean.sum(axis=1), axis=0)
        ind = self.cellparams['group'].apply(lambda x: 'group%d_genemean' % x)

        if self.nproggenes == 0:
            cellgenemean = group_genemean.loc[ind,:].astype(float)
            cellgenemean.index = self.cellparams.index
        else:
            noprogcells = self.cellparams['has_program']==False
            hasprogcells = self.cellparams['has_program']==True

            print('   - Getting mean for activity program carrying cells')
            progcellmean = group_genemean.loc[ind[hasprogcells], :]
            progcellmean.index = ind.index[hasprogcells]
            progcellmean = progcellmean.multiply(1-self.cellparams.loc[hasprogcells, 'program_usage'], axis=0)

            progmean = self.geneparams.loc[:,['prog_genemean']]
            progmean = progmean.div(progmean.sum(axis=0), axis=1)
            progusage = self.cellparams.loc[progcellmean.index, ['program_usage']]
            progusage.columns = ['prog_genemean']
            progcellmean += progusage.dot(progmean.T)
            progcellmean = progcellmean.astype(float)

            print('   - Getting mean for non activity program carrying cells')
            noprogcellmean = group_genemean.loc[ind[noprogcells],:]
            noprogcellmean.index = ind.index[noprogcells]

            cellgenemean = pd.concat([noprogcellmean, progcellmean], axis=0)

            del(progcellmean, noprogcellmean)

            cellgenemean = cellgenemean.reindex(index=self.cellparams.index)

        print('   - Normalizing by cell libsize')
        normfac = (self.cellparams['libsize'] / cellgenemean.sum(axis=1)).values
        for col in cellgenemean.columns:
            cellgenemean[col] = cellgenemean[col].values*normfac
        #cellgenemean = cellgenemean.multiply(normfac, axis=0).astype(float)
        return(cellgenemean)


    def get_gene_params(self):
        '''Sample each genes mean expression from a gamma distribution as
        well as identifying outlier genes with expression drawn from a
        log-normal distribution'''
        basegenemean = np.random.gamma(shape=self.mean_shape,
                                       scale=1./self.mean_rate,
                                       size=self.ngenes)

        is_outlier = np.random.choice([True, False], size=self.ngenes,
                                      p=[self.expoutprob,1-self.expoutprob])
        outlier_ratio = np.ones(shape=self.ngenes)
        outliers = np.random.lognormal(mean=self.expoutloc,
                                       sigma=self.expoutscale,
                                       size=is_outlier.sum())
        outlier_ratio[is_outlier] = outliers
        gene_mean = basegenemean.copy()
        median = np.median(basegenemean)
        gene_mean[is_outlier] = outliers*median
        self.genenames = ['Gene%d' % i for i in range(1, self.ngenes+1)]
        geneparams = pd.DataFrame([basegenemean, is_outlier, outlier_ratio, gene_mean],
                                  index=['BaseGeneMean', 'is_outlier', 'outlier_ratio', 'gene_mean'],
                                 columns=self.genenames).T
        return(geneparams)


    def get_cell_params(self):
        '''Sample cell group identities and library sizes'''
        groupid = self.simulate_groups()
        libsize = np.random.lognormal(mean=self.libloc, sigma=self.libscale,
                                      size=self.init_ncells)
        self.cellnames = ['Cell%d' % i for i in range(1, self.init_ncells+1)]
        cellparams = pd.DataFrame([groupid, libsize],
                                  index=['group', 'libsize'],
                                  columns=self.cellnames).T
        cellparams['group'] = cellparams['group'].astype(int)
        return(cellparams)


    def simulate_program(self):
        ## Simulate the program gene expression
        self.geneparams['prog_gene'] = False
        proggenes = self.geneparams.index[-self.nproggenes:]
        self.geneparams.loc[proggenes, 'prog_gene'] = True
        DEratio = np.random.lognormal(mean=self.progdeloc,
                                          sigma=self.progdescale,
                                          size=self.nproggenes)
        DEratio[DEratio<1] = 1 / DEratio[DEratio<1]
        is_downregulated = np.random.choice([True, False],
                                            size=len(DEratio),
                                            p=[self.progdownprob,
                                            1-self.progdownprob])
        DEratio[is_downregulated] = 1. / DEratio[is_downregulated]
        all_DE_ratio = np.ones(self.ngenes)
        all_DE_ratio[-self.nproggenes:] = DEratio
        prog_mean = self.geneparams['gene_mean']*all_DE_ratio
        self.geneparams['prog_genemean'] = prog_mean

        ## Assign the program to cells
        self.cellparams['has_program'] = False
        if self.proggoups is None:
            ## The program is active in all cell types
            self.proggoups = np.arange(1, self.ngroups+1)

        self.cellparams.loc[:, 'program_usage'] = 0
        for g in self.proggoups:
            groupcells = self.cellparams.index[self.cellparams['group']==g]
            hasprog = np.random.choice([True, False], size=len(groupcells),
                                      p=[self.progcellfrac,
                                      1-self.progcellfrac])
            self.cellparams.loc[groupcells[hasprog], 'has_program'] = True
            usages = np.random.uniform(low=self.minprogusage,
                                       high=self.maxprogusage,
                                       size=len(groupcells[hasprog]))
            self.cellparams.loc[groupcells[hasprog], 'program_usage'] = usages




    def simulate_groups(self):
        '''Sample cell group identities from a categorical distriubtion'''
        groupid = np.random.choice(np.arange(1, self.ngroups+1),
                                       size=self.init_ncells, p=self.groupprob)
        self.groups = np.unique(groupid)
        return(groupid)


    def sim_group_DE(self):
        '''Sample differentially expressed genes and the DE factor for each
        cell-type group'''
        groups = self.cellparams['group'].unique()
        if self.nproggenes>0:
            proggene = self.geneparams['prog_gene'].values
        else:
            proggene = np.array([False]*self.geneparams.shape[0])

        for group in self.groups:
            isDE = np.random.choice([True, False], size=self.ngenes,
                                      p=[self.diffexpprob,1-self.diffexpprob])
            isDE[proggene] = False # Program genes shouldn't be differentially expressed between groups
            DEratio = np.random.lognormal(mean=self.diffexploc,
                                          sigma=self.diffexpscale,
                                          size=isDE.sum())
            DEratio[DEratio<1] = 1 / DEratio[DEratio<1]
            is_downregulated = np.random.choice([True, False],
                                            size=len(DEratio),
                                            p=[self.diffexpdownprob,1-self.diffexpdownprob])
            DEratio[is_downregulated] = 1. / DEratio[is_downregulated]
            all_DE_ratio = np.ones(self.ngenes)
            all_DE_ratio[isDE] = DEratio
            group_mean = self.geneparams['gene_mean']*all_DE_ratio

            deratiocol = 'group%d_DEratio' % group
            groupmeancol = 'group%d_genemean' % group
            self.geneparams[deratiocol] = all_DE_ratio
            self.geneparams[groupmeancol] = group_mean



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sun Jan 19 03:01:35 2020

@author: heavens
"""
import numpy as np
from scipy.special import softmax
from scipy.spatial.distance import cdist
from scipy.stats import multinomial
import pandas as pd
from numpy.random import uniform
from numpy.random import choice
from .opt import valid_neighbourhood_frequency
from .data_op import KL_divergence
from math import factorial
from .random_generator import multinomial_wrapper as mp
from .data_op import DataLoader
from .scsim import scsim
import time
import pickle

def save_simulation(sim,file):
    """Save a Simulator instance to file.
    Args:
        sim: A instance of the Simulator.
        file: The file to save.
    """
    with open(file,'wb+') as f:
        pickle.dump(sim,f)

def load_simulation(file):
    """Load a Simulator instance from file
    Args:
        file: A file path to load the simulator.
    """
    with open(file,'rb') as f:
        instance = pickle.load(f)
    return instance

def get_gene_prior(gene_expression,cell_types,header = 0,):
    """Get the prior parameter (mean and std) of the given gene expression
    matrix and cell type assignment.
    """
    gene_expression = np.asarray(gene_expression)
    gene_mean = []
    gene_std = []
    tags = np.unique(cell_types)
    for tag in tags:
        sub_data = gene_expression[cell_types == tag]
        gene_mean.append(np.mean(sub_data,axis = 0))
        gene_std.append(np.std(sub_data,axis = 0))
    return np.asarray(gene_mean),np.asarray(gene_std)

def get_nf_prior(coordinates,cell_types,threshold = 100):
    distance_matrix = cdist(coordinates,coordinates,'euclidean')
    adjacency = distance_matrix<threshold
    type_tages_all,type_count = np.unique(cell_types,return_counts = True)
    type_indexs = {}
    for tag_id,tag in enumerate(type_tages_all):
        type_indexs[tag] = tag_id
    n_types = len(type_tages_all)
    nb_freq = np.zeros((n_types,n_types))
    n_sample = coordinates.shape[0]
    for i in range(n_sample):
        cell_bags = cell_types[adjacency[i,:]]
        type_tags,counts = np.unique(cell_bags,return_counts = True)
        for j,tag in enumerate(type_tags):
            nb_freq[type_indexs[cell_types[i]],type_indexs[tag]] += counts[j]
    nb_freq += 1e-3 #Add pseudocount
    nb_freq = nb_freq/np.sum(nb_freq,axis = 1,keepdims = True)
    return nb_freq,type_tages_all,type_count

class Simulator():
    """A simulator for generating gene expression and spatio location profile.
    Args:
        sample_n: The number of sample(cell) need to be generated.
        gene_n: The number of genes for each cell.
        cell_type_n: The number of cell types.
        density: The number o fcell in the unit cycle(also it's the average neighbourhood)
        seed: Random seed, default None.
    """
    def __init__(self,
                 sample_n,
                 gene_n,
                 cell_type_n,
                 density,
                 seed = None):
        self.sample_n = sample_n
        self.gene_n = gene_n
        self.cell_n = cell_type_n
        self.seed = seed
        self.target_freq = None
        self.density = None
        self.xrange = None
        self.coor = None
        self.distance_matrix = None
        self.adjacency = None
        self._neighbourhood_frequency = None
        self._neighbourhood_count = None
        self.cell_type_assignment = None
        self.all_types = None
        self.cell_prior = None
    def gen_data_from_real(self,gene_expression,cell_type_list,neighbour_freq_prior):
        gene_mean,gene_std = get_gene_prior(gene_expression,cell_types)
        target_freq = (neighbour_freq_prior+0.1)/np.sum(neighbour_freq_prior+0.1,axis=1,keepdims=True)
        result = valid_neighbourhood_frequency(target_freq)
        self.target_freq = result[0]
        self.gen_parameters(gene_mean_prior = gene_mean[:,:self.gene_n])
        self.gen_coordinate(density = self.density)
        self.assign_cell_type(target_neighbourhood_frequency=self.target_freq)
        
    def gen_parameters(self,
                        gene_mean_prior = None,
                        cell_prior = None,
                        seed = None):
        self.all_types = np.arange(self.cell_n)
        if seed is None:
            np.random.seed(self.seed)
        else:
            np.random.seed(seed)
        self.g_mean = np.exp(np.random.rand(self.cell_n,self.gene_n)+1)
        if gene_mean_prior is not None:
            self.g_mean += gene_mean_prior
        g_cov = np.random.rand(self.cell_n,self.gene_n,self.gene_n)
        self.g_cov = np.asarray([np.dot(x,x.transpose())/self.gene_n for x in g_cov])
        if cell_prior is None:
            cell_prior = np.random.random(self.cell_n)
            self.cell_prior = softmax(cell_prior+1)
        else:
            self.cell_prior = cell_prior/np.sum(cell_prior)
        self._neighbourhood_count = None
        self._neighbourhood_frequency = None
        
    def gen_coordinate(self,
                       density,
                       use_knearest = False,
                       ref_coor = None,
                       rank_criteria = "square"):
        """Random assign the coordinate
        Args:
            density: The sample density, number of samples in unit circle.
            
        """
        self.density = density
        coor_idxs = None
        if ref_coor is None:
            self.xrange = np.sqrt(np.pi*self.sample_n/self.density)
            self.coor = uniform(high = self.xrange,size = (self.sample_n,2))
        else:
            assert len(ref_coor) >= self.sample_n  #changed
            central_point = np.median(ref_coor,axis = 0)
            ref_coor -= central_point
            if rank_criteria == "square":
                coor_criteria = np.max(np.abs(ref_coor),axis = 1,keepdims = False)**2*4
            elif rank_criteria == "euclidean":
                coor_criteria = np.sqrt(ref_coor[:,0]**2+ref_coor[:,1]**2)**2*np.pi
            coor_rank = np.argsort(coor_criteria)
            coor_idxs = coor_rank[:self.sample_n]
            coor = ref_coor[coor_idxs,:]
            effect_area = self.sample_n/density*np.pi
            scale = np.sqrt(effect_area/coor_criteria[coor_rank[self.sample_n-1]])
            self.coor = coor * scale
            self.xrange = max(self.coor[-1,0],self.coor[-1,1])
            self.ref_coor = ref_coor * scale
            
        self.calculate_adjacency(density = density,use_knearest = use_knearest)
        return coor_idxs
                
    def calculate_adjacency(self,density = None,use_knearest = False):
        if not density:
            density = self.density
        self.distance_matrix = cdist(self.coor,self.coor,'euclidean')
        self.adjacency = np.zeros((self.sample_n,self.sample_n),dtype = bool)
        if use_knearest:
            for i,dist in enumerate(self.distance_matrix):
                sort_idx = np.argsort(dist)
                self.adjacency[i,sort_idx[:density+1]] = True
            self.exclude_adjacency = self.adjacency ^ np.eye(self.sample_n).astype(bool)
        else:
            self.adjacency = self.distance_matrix<1
            self.exclude_adjacency = self.adjacency ^ np.eye(self.sample_n).astype(bool)
        
    @property
    def neighbour_frequency(self):
        if self._neighbourhood_frequency is None:
            self._get_neighbourhood_frequency()
        return self._neighbourhood_frequency
    
    @property
    def neighbour_count(self):
        if self._neighbourhood_frequency is None:
            self._get_neighbourhood_count()
        return self._neighbourhood_count
    def _get_neighbourhood_likelihood(self,mn_pdf):
        ll = np.empty((self.cell_n,self.sample_n))
        for i in np.arange(self.cell_n):
            ll[i,:] = mn_pdf[i].logpmf(self._neighbourhood_count)
        return ll
#moved here            
    def _get_neighbourhood_frequency(self,recount_neighbourhood = True):
        """Get the neighbourhood frequency from the cell type assignment or from
        a given neighbourhood count.
        """
        if self._neighbourhood_frequency is None:
            self._neighbourhood_frequency = np.zeros((self.cell_n,self.cell_n))
        if recount_neighbourhood:
            self._get_neighbourhood_count()
        count = self._neighbourhood_count
        for i in range(self.cell_n):
                self._neighbourhood_frequency[i]=np.sum(count[self.cell_type_assignment==i],axis = 0)
        self._neighbourhood_frequency = self._neighbourhood_frequency/np.sum(self._neighbourhood_frequency,axis = 1,keepdims = True)
    
    def _get_neighbourhood_count(self,i = None):
        one_hot = np.zeros((self.sample_n,self.cell_n))
        one_hot[np.arange(self.sample_n),self.cell_type_assignment] = 1
        if i is None:
            self._neighbourhood_count = np.matmul(self.exclude_adjacency,one_hot)
        else:
            self._neighbourhood_count[i] = np.matmul(self.exclude_adjacency[i],one_hot)

    def assign_cell_type(self,
                         target_neighbourhood_frequency,
                         tol = 1e-2,
                         max_iter = 40000,     ###changed
                         use_exist_assignment = False,
                         method = None,
                         soft_factor = None,
                         annealing = False,
                         initial_temperature = 20,
                         half_decay = 100,
                         local_criteria = True,
                         ref_assignment = None):
        """Generate cell type assignment iteratively from a given neighbourhood
        frequency, require the coordinates being generated first by calling
        self.gen_coordinate.
        Args:
            target_neighbourhood_frequency: A n-by-n target neighbourhood freuency matrix,
                n is the number of cell type, and ith row is the neighbourhood 
                frequency of ith cell, target_neighbourhood_frequency[i].
            tol: The error tolerance.
            max_iter: Max iteraton.
            method: Default is 'swap', can be 'assign-neighbour'
            annealing: Default is False. If annealing is used when the method is Metropolis-swap.
            initial_temperature: The initial temeprature to used in the annealing algorithm.
            half_decay: The steps of the half decay temperature.
            local_criteria: If calculate the swap probabiltiy based on local change.
        """
        self.target_freq = target_neighbourhood_frequency
        self.log_target_freq = np.log(self.target_freq)
        if not use_exist_assignment:
            self.cell_type_assignment = choice(np.arange(self.cell_n),
                                    size = self.sample_n,
                                    p = self.cell_prior).astype(int)
        self._get_neighbourhood_frequency()
        error = np.linalg.norm(self._neighbourhood_frequency-target_neighbourhood_frequency)
        print("    0 iteration, error %.2f"%(error))
        iter_n = 0
        perm = np.arange(self.sample_n)
        cell_types = np.arange(self.cell_n)
        error_record = []
        mps = []
        for i in np.arange(self.cell_n):
            mps.append(mp(target_neighbourhood_frequency[i]))
        while error>tol and iter_n<max_iter:
            
            print()#changed
            if method=='assign-neighbour':
                np.random.shuffle(perm)
                for i in perm:
                    i_type = self.cell_type_assignment[i]
                    mask = np.copy(self.adjacency[i])
                    mask[i] = False #Exclude the self count.
                    neighbour_n = np.sum(mask)
                    self.cell_type_assignment[i] = np.random.choice(cell_types,
                                                                    size = 1,
                                                                    p = self.cell_prior)[0]
                    reasign_type = np.random.choice(cell_types,
                                                    size = neighbour_n,
                                                    p = target_neighbourhood_frequency[i_type])
                    self.cell_type_assignment[mask] = reasign_type
                    iter_n+=1
            elif method is None or method=="Metropolis-swap":
                for i in np.arange(100):
                    swap_cluster = np.random.choice(np.arange(self.cell_n),2,replace = False)
                    cell_idxs = np.arange(self.sample_n)
                    label = self.cell_type_assignment
                    i = np.random.choice(cell_idxs[label==swap_cluster[0]],1)[0]
                    j = np.random.choice(cell_idxs[label==swap_cluster[1]],1)[0]
                    nb_indexs = np.where(np.logical_or(self.exclude_adjacency[i] ,self.exclude_adjacency[j]))[0]
                    dist_before = 0
                    for k, freq in enumerate(self._neighbourhood_frequency):
                        dist_before += KL_divergence(target_neighbourhood_frequency[k],freq)
                    label[j],label[i] = label[i],label[j]
                    self._get_neighbourhood_count(nb_indexs)
                    # Update the neighbourhood count for the neighbourhood
                    # of swap pairs only.
                    self._get_neighbourhood_frequency(recount_neighbourhood = False)
                    dist_after = 0
                    for k, freq in enumerate(self._neighbourhood_frequency):
                        dist_after += KL_divergence(target_neighbourhood_frequency[k],freq)
                    swap_p = np.random.uniform()
                    temperature = initial_temperature*0.5**(iter_n/half_decay)
                    if annealing:
                        swap = temperature*np.log(swap_p)
                    else:
                        swap = 0
                    if swap>(dist_before - dist_after):
                        #Reject the swap, so swap back.
                        label[j],label[i] = label[i],label[j]
                        self._get_neighbourhood_count(nb_indexs)
                        self._get_neighbourhood_frequency(recount_neighbourhood = False)
                    iter_n+=1
            elif method=="Gibbs-sampling":
                np.random.shuffle(perm)
                ll = self._get_neighbourhood_likelihood(mps)
                ll = np.choose(self.cell_type_assignment,ll)
                perm = np.argsort(ll)
                for i in perm:
                    mask = np.copy(self.exclude_adjacency[i])
                    assign_prob = self._assign_probability(i,
                                                           self._neighbourhood_count[i],
                                                           mask,
                                                           mps,
                                                           self.cell_prior,
                                                           s =soft_factor)
                    print(assign_prob)
                    assign_cell_type = np.random.choice(cell_types,
                                                        size = 1,
                                                        p = assign_prob)[0]
                    self.cell_type_assignment[i] = assign_cell_type
                    self._get_neighbourhood_count(i=mask)
                    iter_n+=1
            elif method=="Direct-assignment":
                self.cell_type_assignment = ref_assignment
                self._get_neighbourhood_count()
                self._get_neighbourhood_frequency(recount_neighbourhood = False)
                return None
                
            if iter_n %100 == 0:
                #Update error every 100 iterations.
                error = np.linalg.norm(self._neighbourhood_frequency-target_neighbourhood_frequency)
            if iter_n%1000 == 0:
                self._get_neighbourhood_frequency(recount_neighbourhood = False)
                error = np.linalg.norm(self._neighbourhood_frequency-target_neighbourhood_frequency)
                print("%5d iteration, error %.2f"%(iter_n,error))
                error_record.append(error)
                klds = np.empty(self.cell_n)
                for i, freq in enumerate(self._neighbourhood_frequency):
                    klds[i] = KL_divergence(target_neighbourhood_frequency[i],freq)
#                print("%d iteration, error %.2f"%(iter_n,error))
#                print("KL divergence %s"%(",".join([str(round(x,3)) for x in klds])))
#                print(np.unique(self.cell_type_assignment,return_counts = True))
        return error_record
    
    def _assign_probability(self,
                            cell_index,
                            neighbourhood_count,
                            neighbourhood_mask,
                            target_neighbourhood_frequency_pdf,
                            prior,
                            s=None):
        """Calculate the posterior probability given the neighbourhood cell type.
        Args:
            neighbourhood_count: A C vector indicate the count of the current cell.
            neighbourhood_mask: A length N boolean vector indicate the neighbourhood
                of current cell.
            target_neighbourhood_frequency_pdf:A length C list contain the
                multinomial distribution of target frequency of each cell type.
            prior: A length C vector indicate the prior probability of each cell type.
            s: Soft factor, if None, using the density.
        """
        neighbourhood_matrix = self._neighbourhood_count[neighbourhood_mask]
        neighbourhood_cell_type = self.cell_type_assignment[neighbourhood_mask]
        posterior = np.zeros(self.cell_n)
        current_cell_type = self.cell_type_assignment[cell_index]
        if s is None:
            s = self.density
        nb_copy = neighbourhood_matrix.copy()
        nb_copy[:,current_cell_type] -=1
        for i in np.arange(self.cell_n):
            posterior[i] = np.sum([self.log_target_freq[cl,i]-np.log(nb_copy[l,i]+1) 
                                  for l,cl in enumerate(neighbourhood_cell_type)])+\
                           np.sum([self.log_target_freq[i,j]*neighbourhood_count[j] 
                                  for j in np.arange(self.cell_n)])
        assign_prob = posterior/s
        assign_prob = softmax(assign_prob)
        return assign_prob
    
    def _get_neighbourhood_likelihood(self,mn_pdf):
        ll = np.empty((self.cell_n,self.sample_n))
        for i in np.arange(self.cell_n):
            ll[i,:] = mn_pdf[i].logpmf(self._neighbourhood_count)
        return ll
            
    def _get_neighbourhood_frequency(self,recount_neighbourhood = True):
        """Get the neighbourhood frequency from the cell type assignment or from
        a given neighbourhood count.
        """
        if self._neighbourhood_frequency is None:
            self._neighbourhood_frequency = np.zeros((self.cell_n,self.cell_n))
        if recount_neighbourhood:
            self._get_neighbourhood_count()
        count = self._neighbourhood_count
        for i in range(self.cell_n):
                self._neighbourhood_frequency[i]=np.sum(count[self.cell_type_assignment==i],axis = 0)
        self._neighbourhood_frequency = self._neighbourhood_frequency/np.sum(self._neighbourhood_frequency,axis = 1,keepdims = True)
    
    def _get_neighbourhood_count(self,i = None):
        one_hot = np.zeros((self.sample_n,self.cell_n))
        one_hot[np.arange(self.sample_n),self.cell_type_assignment] = 1
        if i is None:
            self._neighbourhood_count = np.matmul(self.exclude_adjacency,one_hot)
        else:
            self._neighbourhood_count[i] = np.matmul(self.exclude_adjacency[i],one_hot)

    
    def gen_expression(self,
                       zeroing = True, 
                       seed = None,
                       drop_rate = None):
        """Generate gene expression, need to call assign_cell_type first.
        """
        gene_expression = np.empty((self.sample_n,self.gene_n),dtype = float)
        if seed is None:
            np.random.seed(self.seed)
        else:
            np.random.seed(seed)
        if drop_rate is not None:
            assert len(drop_rate) ==  self.gene_n
            mask = np.empty((self.sample_n,self.gene_n),dtype = bool)
            for i in np.arange(self.gene_n):
                mask[:,i] = np.random.choice([True,False],size = (self.sample_n),p = [drop_rate[i],1-drop_rate[i]])
        for i in range(self.cell_n):
            cell_type_mask = self.cell_type_assignment == i
            count = np.sum(cell_type_mask)
            current_expression = np.random.multivariate_normal(mean = self.g_mean[i],
                                                               cov = self.g_cov[i],
                                                               size = count)
            if zeroing:
                current_expression[current_expression<0] = 0
            if drop_rate is not None:
                current_expression[mask[cell_type_mask,:]] = 0
            gene_expression[cell_type_mask,:] = current_expression
        return np.asarray(gene_expression),self.cell_type_assignment,self._neighbourhood_count
    def gen_expression_splatter(self,seed=None):
        if seed is None:
            np.random.seed(self.seed)
        else:
            np.random.seed(seed)
        celltype = self.cell_type_assignment
        n_g = self.gene_n
        sample_n = self.sample_n
        celltype_n = self.cell_n
        deval = 1
        K = celltype_n
        doubletfrac = 0
        ncells = sample_n
        ngenes=n_g
        nproggenes = 400
        ndoublets=int(doubletfrac*ncells)
        groupid = celltype
        deloc=deval
        progdeloc=deval
        descale=1.0
        progcellfrac = .35
        deprob = .025
        nproggroups = K #Enable program genes in all the groups
        proggroups = list(range(1, nproggroups+1))
        print("Simulation using splatter.")
        simulator = scsim(ngenes=ngenes, ncells=ncells, ngroups=K,groupid = groupid, libloc=7.64, libscale=0.78,
                     mean_rate=7.68,mean_shape=0.34, expoutprob=0.00286,
                     expoutloc=6.15, expoutscale=0.49,
                     diffexpprob=deprob, diffexpdownprob=0., diffexploc=deloc, diffexpscale=descale,
                     bcv_dispersion=0.448, bcv_dof=22.087, ndoublets=ndoublets,
                     nproggenes=nproggenes, progdownprob=0., progdeloc=progdeloc,
                     progdescale=descale, progcellfrac=progcellfrac, proggoups=proggroups,
                     minprogusage=.1, maxprogusage=.7, seed=seed)
        start_time = time.time()
        simulator.simulate()
        end_time = time.time()
        print("Elapsing time is %.2f"%(end_time - start_time))
        self.gene_expression = np.asarray(simulator.counts)
        return self.gene_expression,self.cell_type_assignment,self._neighbourhood_count
class SimDataLoader(DataLoader):
    def __init__(self,
                 gene_expression,
                 cell_neighbour,
                 cell_type_assignment):
        """Class for loading the data.
        Input Args:
            gene_expression: A N-by-M matrix contain the gene expression for N samples and M genes.
            cell_neighbour: A N-by-C matrix contain the neighbourhood count for N samples and C cell types.
            cell_type_assignment: If the dataset is not for evaluation, then a length N vector indicate the
                cell_type_assignment need to be provided.
        """
        super().__init__(xs = (gene_expression,cell_neighbour),
                         y = cell_type_assignment,
                         for_eval = False)


if __name__ == "__main__":
    # ### Hyper parameter setting
    # sample_n = 1000 #Number of samples
    # n_g = 100 #Number of genes
    # n_c = 10 #Number of cell type
    # density = 20 #The average number of neighbour for each cells.
    # threshold_distance = 1 # The threshold distance of neighbourhood.
    # gene_col = np.arange(9,164)
    # coor_col = [5,6]
    # header = 1
    # data_f = "/home/heavens/CMU/FISH_Clustering/FICT/example_data2/aau5324_Moffitt_Table-S7.xlsx"
    # model_f = "/home/heavens/CMU/FISH_Clustering/test_sim1"
    
    # ### Data preprocessing
    # data = pd.read_excel(data_f,header = header)
    # gene_expression = data.iloc[:,gene_col]
    # cell_types = data['Cell_class']
    # type_tags = np.unique(cell_types)
    # coordinates = data.iloc[:,coor_col]
    
    # ### Choose only the n_c type cells
    # if len(type_tags)<n_c:
    #     raise ValueError("Only %d cell types presented in the dataset, but require %d, reduce the number of cell type assigned."%(len(type_tags),n_c))
    # mask = np.asarray([False]*len(cell_types))
    # for tag in type_tags[:n_c]:
    #     mask = np.logical_or(mask,cell_types==tag)
    # gene_expression = gene_expression[mask]
    # cell_types = np.asarray(cell_types[mask])
    # coordinates = np.asarray(coordinates[mask])
    
    # ### Generate prior from the given dataset.
    # gene_mean,gene_std = get_gene_prior(gene_expression,cell_types)
    # drop_rate = np.sum(gene_expression==0,axis = 0)/(gene_expression.shape[0])
    # neighbour_freq_prior,tags,type_count = get_nf_prior(coordinates,cell_types)
    # type_prior = type_count/np.sum(type_count)
    # target_freq = (neighbour_freq_prior+0.1)/np.sum(neighbour_freq_prior+0.1,axis=1,keepdims=True)
    # result = valid_neighbourhood_frequency(target_freq)
    # target_freq = result[0]
    
    # ### Generate simulation dataset and load
    # sim = Simulator(sample_n,n_g,n_c,density)
    # sim.gen_parameters(gene_mean_prior = gene_mean[:,:n_g])
    # sim.gen_coordinate(density = density)
    # sim.assign_cell_type(target_neighbourhood_frequency=target_freq, method = "assign-neighbour")
    # gene_expression,cell_type,cell_neighbour = sim.gen_expression(drop_rate = drop_rate[:n_g])
    # save_simulation(sim,model_f)
    # sim2 = load_simulation(model_f)
    
    ### Generate simulation dataset and load
    sim = Simulator(sample_n,n_g,n_c,density)
    sim.gen_parameters(gene_mean_prior = gene_mean[:,:n_g])
    sim.gen_coordinate(density = density)
    sim.assign_cell_type(target_neighbourhood_frequency=target_freq, method = "assign-neighbour")
    gene_expression,cell_type,cell_neighbour = sim.gen_expression(drop_rate = drop_rate[:n_g])
    save_simulation(sim,model_f)
    sim2 = load_simulation(model_f)




import io
import os
from setuptools import setup, find_packages

here = os.path.abspath(os.path.dirname(__file__))

try:
    with io.open(os.path.join(here, 'README.md'), encoding='utf-8') as f:
        long_description = '\n' + f.read()
except FileNotFoundError:
    long_description = DESCRIPTION


# Package meta-data.
NAME = 'st-spider'
DESCRIPTION = 'A tools to simulate spatial transcriptomics data.'
EMAIL = '599568651@qq.com'
URL="https://github.com/YANG-ERA/Spider/tree/main"
AUTHOR ='Jiyuan Yang'
VERSION = '0.1.9'

setup(
    name=NAME,
    version=VERSION,
    author=AUTHOR,
    author_email=EMAIL,
	license='MIT',
    description=DESCRIPTION,
	url=URL,
    long_description_content_type="text/markdown",
    long_description=long_description,
    packages=find_packages(),
    install_requires=["anndata", "matplotlib", "numba", "numpy",
                  "pandas", "scanpy", "scikit_learn",
                  "scipy", "seaborn", "squidpy", "torch"]
)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sun Feb  9 00:03:44 2020

@author: heavens
"""

from scipy.optimize import minimize
import numpy as np
from numpy.random import uniform

def symmetric_loss(D,M,l=0):
    loss = np.linalg.norm(M*D - np.transpose(M*D)) + l*np.abs(np.sum(D)-1)
    return loss

def normalize(M,axis = 1):
    M = M/np.sum(M,axis = axis,keepdims = True)
    return M

def constrain(D):
    return np.sum(D)-1

def reconstruct(init_M,Diagonal):
    count = init_M*Diagonal
    count = (count + np.transpose(count))/2
    return count

def valid_neighbourhood_frequency(init_frequency,norm_axis = 1):
    init_frequency = normalize(init_frequency,axis = norm_axis)
    n = init_frequency.shape[0]
    D = [1/n]*n
    if norm_axis == 1:
        init_frequency = np.transpose(init_frequency)
    eq_cons = {'type': 'eq',
               'fun':constrain}
    cons = [eq_cons]
    for i in range(n):
        cons.append({'type':'ineq',
                 'fun':lambda x:x[i]})
    opt = minimize(symmetric_loss,
                   D,
                   method = 'trust-constr',
                   args = init_frequency,
                   constraints = cons,
                   tol = 1e-5)
    new_count = reconstruct(init_frequency,opt.x)
    if norm_axis == 1:
        init_frequency = np.transpose(init_frequency)
    return normalize(new_count,axis = norm_axis),opt
    
if __name__ == "__main__":
#    count_matrix = np.asarray([[2,4,8],[4,3,7],[8,6,5]])
#    test_matrix = normalize(count_matrix)
#    D0 = [0.5,0.2,0.3]
#    result = minimize(symmetric_loss,
#                      D0,
#                      args = test_matrix,
#                      method='BFGS',
#                      tol = 1e-3)
#    print(result.fun)
#    print("Count matrix:")
#    print(test_matrix*result.x)
#    print("Given prob matrix:")
#    print(test_matrix)
#    print("Recorvered matrix:")
#    print(reconstruct(test_matrix,result.x))
    text_f = "/home/heavens/CMU/FISH_Clustering/FICT/out.txt"
    test_matrix2 = np.random.rand(10,10)
    print("Given prob matrix:")
    print(normalize(test_matrix2))
    print("Optimized nearest valid matrix:")
    result,opt = valid_neighbourhood_frequency(test_matrix2)
    print(normalize(result))
    with open(text_f,'w') as f:
        for i in range(10):
            f.write(' '.join([str(x) for x in result[i,:]]))
            f.write("\n")
            f.write(' '.join([str(x) for x in normalize(test_matrix2)[i,:]]))
            f.write("\n")
            f.write("%"*10)
            f.write("\n")


from .run_scsim import *
from .scsim import *
from .sim_expr import *
from .simulate_10X import *
from .Annealing import *
from .sim_naive import *

from .joint_simulator import *
from .opt import *
from .data_op import *
from .random_generator import *





#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Mar  2 00:44:40 2020

@author: heavens
"""

from scipy.stats import multinomial
from scipy.stats import dirichlet
import numpy as np
from math import gamma

class multinomial_wrapper(object):
    """A wrapper for the scipy multinomial distrbution allow a dynmic assigned n.
    """
    def __init__(self,p):
        self.pdf = multinomial
        self.p = p
    def pmf(self,count_batch):
        batch_shape = count_batch.shape
        if len(batch_shape) == 1:
            n = np.sum(count_batch)
            f = self.pdf(n=n, p = self.p)
            return f.pmf(count_batch)
        results = []
        for count in count_batch:
            n = np.sum(count)
            f = self.pdf(n=n,p=self.p)
            results.append(f.pmf(count))
        return np.asarray(results)
    
    def logpmf(self,count_batch):
        batch_shape = count_batch.shape
        if len(batch_shape) == 1:
            n = np.sum(count_batch)
            f = self.pdf(n=n, p = self.p)
            return f.logpmf(count_batch)
        results = []
        for count in count_batch:
            n = np.sum(count)
            f = self.pdf(n=n,p=self.p)
            results.append(f.logpmf(count))
        return np.asarray(results)

class continuous_multinomial(object):
    """A implementation of the analytic ontinuation of multinomial distribution.
    """
    def __init__(self,p):
        self.p = p
        self.log_p = np.log(p)
        
    def _func(self,x):
        return np.log(gamma(sum(x)+1))-np.sum([np.log(gamma(xi+1)) for xi in x])+np.dot(self.log_p,x)
    
    def pmf(self,count_batch):
        batch_shape = count_batch.shape
        if len(batch_shape) == 1:
            count_batch = [count_batch]
        return np.asarray([np.exp(self._func(x)) for x in count_batch])
    
    def logpmf(self,count_batch):
        batch_shape = count_batch.shape
        if len(batch_shape) == 1:
            count_batch = [count_batch]
        return np.asarray([self._func(x) for x in count_batch])